import abc

from langchain_core.language_models import BaseChatModel


class Prompt:
    def __init__(self, prompt: str):
        self.prompt = prompt

    @abc.abstractmethod
    def invoke(self, model: BaseChatModel):
        pass



from langchain_core.language_models import BaseChatModel

from ceo.action.action import Action
from ceo.prompt.prompt import Prompt


class SchedulerPrompt(Prompt):
    def __init__(self, query: str, actions: list[Action]):
        self.actions = actions
        prompt = dict()
        for action in self.actions:
            prompt[action.name] = str(action)
        prompt = ('Precondition: Below are the tools you can use (you can only use the following tools). '
                  f'Now there is a user query: "{query}"\n'
                  'Task: What you need to do is to plan your workflow based on the tools you have to accomplish the user query '
                  '(make sure your use of tools is appropriate)\n'
                  'Output format: [{tool1.name}, {tool2.name}, ...] sequential and well-organized with no additional redundant information\n'
                  'Example output: [do_step_one, do_step_two, do_step_three]\n'
                  f'Tools: {prompt}\n')
        super().__init__(prompt)

    def invoke(self, model: BaseChatModel) -> list[Action]:
        results = model.invoke(self.prompt)
        results = results.content[1:-1].split(',')
        _fin_results = list()
        for _a_result in results:
            for action in self.actions:
                if action.name == _a_result.strip():
                    _fin_results.append(action)
        return _fin_results



import json

from langchain_core.language_models import BaseChatModel

from ceo.action.action import Action
from ceo.prompt.prompt import Prompt


class AnalyserPrompt(Prompt):
    def __init__(self, query: str, prev_results: list, action: Action):
        self.action = action
        prev_results_str = str()
        prompt = str()
        for i, result in enumerate(prev_results):
            prev_results_str += f'step{i}:{result};'
        prompt += ('Precondition: Below is the tool you can use (you can only use this tool). '
                   f'Now there is a user query: "{query}"\n'
                   'Task: What you need to do is to generate values of parameters of the function\n'
                   'Output format: {param1_name:param1_value, param2_name:param2_value, ...} '
                   '(params are given below, values are gonna be generated by you, '
                   'strictly abide by the parameter name and type)\n'
                   'Example output: {"param1": "value1", "param2": "value2"} '
                   '(With not any redundant information, only a json)\n"}'
                   f'Tool: {action}\n')
        if len(prev_results_str) != 0:
            prompt += (f'Generate response according to previous actions performed by you.\n'
                       f'Previous actions: {prev_results_str}\n')
        super().__init__(prompt)

    def invoke(self, model: BaseChatModel) -> tuple[Action, dict]:
        result = model.invoke(self.prompt).content
        param = json.loads(result)
        return self.action, param



from typing import Callable

from langchain_core.language_models import BaseChatModel

from ceo.action.action import Action
from ceo.prompt import SchedulerPrompt, AnalyserPrompt, ExecutorPrompt, IntrospectionPrompt


class Agent:
    def __init__(self, functions: list[Callable], model: BaseChatModel):
        self.actions = list()
        self.prev_results = list()
        self.schedule = list()
        self.act_count = 0
        self.model = model
        for function in functions:
            self.actions.append(Action(function))

    def plan(self, query: str) -> list:
        scheduling = SchedulerPrompt(query=query, actions=self.actions)
        self.schedule = scheduling.invoke(self.model)
        return self.schedule

    def renew(self):
        self.prev_results = list()
        self.schedule = list()
        self.act_count = 0

    def step_quiet(self, query: str) -> list:
        if self.act_count < len(self.schedule):
            analysing = AnalyserPrompt(
                query=query,
                prev_results=self.prev_results,
                action=self.schedule[self.act_count]
            )
            action, params = analysing.invoke(self.model)
            executing = ExecutorPrompt(params=params, action=action)
            self.prev_results.append(executing.invoke(model=self.model))
            self.act_count += 1
            return self.prev_results
        self.renew()
        return self.prev_results

    def just_do_it(self, query: str) -> str | None:
        if not self.plan(query=query):
            return None
        for act_count in range(len(self.schedule)):
            self.step_quiet(query=query)
        response = IntrospectionPrompt(query=query, prev_results=self.prev_results).invoke(self.model)
        self.renew()
        return response
